#include "xufspsxc.h"

XUfsPsxc_Config XUfsPsxc_ConfigTable[] __attribute__ ((section (".drvcfg_sec"))) = {

	{
		"amd,versal2-ufs", /* compatible */
		0xf10b0000, /* reg */
		0x11e19751, /* xlnx,clk-freq-hz */
		0x18cba80, /* xlnx,cfg-clk-freq-hz */
		0x18cba80, /* xlnx,ref-pad-clk-freq-hz */
		0x0, /* xlnx,is-cache-coherent */
		0x40ea, /* interrupts */
		0xe2000000 /* interrupt-parent */
	},
	 {
		 NULL
	}
};